/*****************************************************************************
* File....: ARJDIR.PRG                                                       *
* By......: David A Pearson                                                  *
* Function: Provides a set of functions for extracting the directory of an   *
* ........: ARJ archive file.                                                *
* Notes...: Compile with switches /n/w                                       *
*****************************************************************************/

#include "misc.ch"
#include "fileio.ch"
#include "arjdir.ch"

#define HEADER_ID          60000       // Header ID for an ARJ archive.
#define BASIC_HDR_SIZE     34          // Basic size of a header.
#define FILENAME_MAX       512         // Max length if a filename.
#define COMMENT_MAX        2048        // Max length of a comment.
#define EXTFILE_FLAG       8           // Include start pos field.

// In-line replacement functions for extracting information from the file
// header.

#xtranslate HeaderId(<cHeader>)   => bin2w(left(<cHeader>,2))
#xtranslate HeaderSize(<cHeader>) => bin2w(substr(<cHeader>,3,1))
#xtranslate HostOS(<cHeader>)     => asc(substr(<cHeader>,8,1))
#xtranslate Flags(<cHeader>)      => asc(substr(<cHeader>,9,1))
#xtranslate CompMethod(<cHeader>) => asc(substr(<cHeader>,10,1))
#xtranslate FileType(<cHeader>)   => asc(substr(<cHeader>,11,1))
#xtranslate DateStamp(<cHeader>)  => DecodeDate(substr(<cHeader>,13,4))
#xtranslate TimeStamp(<cHeader>)  => DecodeTime(substr(<cHeader>,13,4))
#xtranslate CompSize(<cHeader>)   => bin2l(substr(<cHeader>,17,4))
#xtranslate OrigSize(<cHeader>)   => bin2l(substr(<cHeader>,21,4))

static nNextHeaderPos := 0       // Position of next header in file.
static lBadArjFile    := FALSE   // Found a problem with the ARJ file.
static nArjDirError   := 0       // Error code of last ARJ read.

/*****************************************************************************
* Function: ArjDirectory()                                                   *
* Syntax..: ArjDirectory(<cArjName>) --> aDirectory                          *
* Usage...: Returns the directory details for the ARJ file.                  *
* By......: David A Pearson                                                  *
*****************************************************************************/

function ArjDirectory(cArjFile)
local aDirectory := {},;
      aFile      := {},;
      nArjFile   := 0
lBadArjFile  := FALSE
nArjDirError := ARJ_ERR_ALL_OK
if !empty(alltrim(cArjFile := Default(cArjFile,NULL)))
   if file(cArjFile)
      if (nArjFile := fopen(cArjFile,FO_READ)) != F_ERROR
         if !(lBadArjFile := !ReadMainHeader(nArjFile))
            do forever
               aFile := ReadNextHeader(nArjFile)
               if nNextHeaderPos != 0 .and. !lBadArjFile
                  aadd(aDirectory,aFile)
               else
                  exit
               endif
            enddo
         endif
         fclose(nArjFile)
      else
         nArjDirError := ARJ_ERR_DOS
      endif
   else
      nArjDirError := ARJ_ERR_NO_FILE
   endif
else
   nArjDirError := ARJ_ERR_BAD_PARAM
endif
return(if(lBadArjFile,{},aDirectory))

/*****************************************************************************
* Function: ArjDirErr()                                                      *
* Syntax..: ArjDirErr() --> nErrorCode                                       *
* Usage...: Returns result code of the last ARJ file read.                   *
* By......: David A Pearson                                                  *
*****************************************************************************/

function ArjDirErr()
return(nArjDirError)

/*****************************************************************************
* Function: ReadMainHeader()                                                 *
* Syntax..: RedaMainHeader(<nArjFile>) --> lIsArjFile                        *
* Usage...: Skips the main archive header checking that this is an ARJ       *
* ........: archive.                                                         *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function ReadMainHeader(nArjFile)
local cBuffer  := space(BASIC_HDR_SIZE),;
      lIsArj   := FALSE                ,;
      nLastPos := 0
if fread(nArjFile,@cBuffer,BASIC_HDR_SIZE) == BASIC_HDR_SIZE
   if (lIsArj := (bin2w(left(cBuffer,2)) == HEADER_ID))
      ReadName(nArjFile)
      ReadComment(nArjFile)
      FindNextHeader(nArjFile,0)
   else
      nArjDirError := ARJ_ERR_NOT_ARJ_FILE
   endif
else
   nArjDirError := ARJ_ERR_BAD_HDR_LEN
endif
return(lIsArj)

/*****************************************************************************
* Function: ReadNextHeader()                                                 *
* Syntax..: RedaNextHeader(<nArjFile>) --> aHeaderInfo                       *
* Usage...: Reads the file header details.                                   *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function ReadNextHeader(nArjFile)
local aHeadInfo := array(ARJ_PACKET_SIZE),;
      cBuffer   := space(BASIC_HDR_SIZE) ,;
      nLastPos  := 0
fseek(nArjFile,nNextHeaderPos,FS_SET)
if !(lBadArjFile := (FTell(nArjFile) != nNextHeaderPos))
   if !(lBadArjFile := (fread(nArjFile,@cBuffer,BASIC_HDR_SIZE) != BASIC_HDR_SIZE))
      if !(lBadArjFile := (HeaderId(cBuffer) != HEADER_ID))
         if HeaderSize(cBuffer) != 0
            aHeadInfo[ARJ_HOST_OS]     := HostOS(cBuffer)
            aHeadInfo[ARJ_COMP_METHOD] := CompMethod(cBuffer)
            aHeadInfo[ARJ_FILE_TYPE]   := FileType(cBuffer)
            aHeadInfo[ARJ_DATE]        := DateStamp(cBuffer)
            aHeadInfo[ARJ_TIME]        := TimeStamp(cBuffer)
            aHeadInfo[ARJ_COMP_SIZE]   := CompSize(cBuffer)
            aHeadInfo[ARJ_ORIG_SIZE]   := OrigSize(cBuffer)
            if And(Flags(cBuffer),EXTFILE_FLAG) != 0
               cBuffer := space(4)
               fread(nArjFile,@cBuffer,4)
            endif
            aHeadInfo[ARJ_FNAME]   := ReadName(nArjFile)
            aHeadInfo[ARJ_COMMENT] := ReadComment(nArjFile)
            FindNextHeader(nArjFile,CompSize(cBuffer))
         else
            lBadArjFile    := FALSE
            nNextHeaderPos := 0
         endif
      else
         lBadArjFile  := TRUE
         nArjDirError := ARJ_ERR_NOT_ARJ_FILE
      endif
   else
      if HeaderSize(cBuffer) == 0
         lBadArjFile    := FALSE
         nNextHeaderPos := 0
      else
         lBadArjFile  := TRUE
         nArjDirError := ARJ_FILE_CORRUPT
      endif
   endif
else
   nArjDirError := ARJ_ERR_LOST_HEADER
   lBadArjFile  := TRUE
endif
return(aHeadInfo)

/*****************************************************************************
* Function: ReadName()                                                       *
* Syntax..: ReadName(<nArjFile>) --> cFileName                               *
* Usage...: Reads the file name from the header.                             *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function ReadName(nArjFile)
local nLastPos := 0,;
      cBuffer  := NULL
nLastPos := FTell(nArjFile)
cBuffer := freadstr(nArjFile,FILENAME_MAX)
fseek(nArjFile,nLastPos+len(cBuffer),FS_SET)
return(cBuffer)

/*****************************************************************************
* Function: ReadComment()                                                    *
* Syntax..: ReadComment(<nArjFile>) --> cComment                             *
* Usage...: Reads the comment from the file header.                          *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function ReadComment(nArjFile)
local nLastPos := 0,;
      cBuffer  := NULL
fseek(nArjFile,(nLastPos := FTell(nArjFile))+1,FS_SET)
cBuffer := freadstr(nArjFile,COMMENT_MAX)
fseek(nArjFile,nLastPos+len(cBuffer),FS_SET)
return(cBuffer)

/*****************************************************************************
* Function: FindNextHeader()                                                 *
* Syntax..: FindNextHeader(<nArjFile>,<nFileSize>) --> NIL                   *
* Usage...: Reads the comment from the file header.                          *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function FindNextHeader(nArjFile,nFileSize)
local cBuffer := space(8)
fread(nArjFile,@cBuffer,8)
nNextHeaderPos := FTell(nArjFile)+nFileSize
return(NIL)

/*****************************************************************************
* Function: DecodeDate()                                                     *
* Syntax..: DecodeDate(<cDateTime>) --> dDate                                *
* Usage...: Gets the date from the date/time value.                          *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function DecodeDate(cDateTime)
local nDate  := int(Bin2UL(cDateTime) / 65536),;
      nYear  := 0                             ,;
      nMonth := 0                             ,;
      nDay   := 0
nYear  := int(nDate / 512) + 1980
nMonth := int(int(nDate / 32) % 16)
nDay   := int(nDate % 32)
return(SToD(alltrim(str(nYear))+padl(nMonth,2,"0")+padl(nDay,2,"0")))

/*****************************************************************************
* Function: DecodeTime()                                                     *
* Syntax..: DecodeTime(<cDateTime>) --> cTime                                *
* Usage...: Gets the time from the date/time value.                          *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function DecodeTime(cDateTime)
local nTime   := int(Bin2UL(cDateTime) % 65536),;
      nHour   := 0                             ,;
      nMinute := 0                             ,;
      nSecond := 0
nHour   := int(nTime / 2048)
nMinute := int(int(nTime / 32) % 64)
nSecond := int(nTime % 32)*2
return(padl(nHour,2,"0")+":"+padl(nMinute,2,"0")+":"+padl(nSecond,2,"0"))

/*****************************************************************************
* Function: Bin2UL()                                                         *
* Syntax..: Bin2UL(<cBin>) --> nUnsignedLong                                 *
* Usage...: Converts a four byte string into an unsigned long value.         *
* By......: David A Pearson                                                  *
*****************************************************************************/

static function Bin2UL(cBin)
return((asc(right(cBin,1))*16777216)+(asc(substr(cBin,3,1))*65536)+;
       (asc(substr(cBin,2,1))*256)+asc(left(cBin,1)))
